{-# LANGUAGE MultiParamTypeClasses #-}

{-| Module  : FiniteCategories
Description : Category of cones and category of cocones of a diagram.
Copyright   : Guillaume Sabbagh 2021
License     : GPL-3
Maintainer  : guillaumesabbagh@protonmail.com
Stability   : experimental
Portability : portable

A cone is an object in the comma category (/D/|1_/F/) where /D/ is the diagonal functor and 1_/F/ is the diagram that selects the diagram of interest in the functor category.

A cocone is an object in the comma category (1_/F/|/D/).
-}

module ConeCategory.ConeCategory
(
    -- * Cone related functions and types.
    Cone,
    ConeMorphism,
    ConeCategory,
    apex,
    coneToNaturalTransformation,
    naturalTransformationToCone,
    mkConeCategory,
    conesOfApex,
    terminalObjects,
    limits,
    -- * Cocone related functions and types.
    Cocone,
    CoconeMorphism,
    CoconeCategory,
    nadir,
    coconeToNaturalTransformation,
    naturalTransformationToCocone,
    mkCoconeCategory,
    coconesOfNadir,
    initialObjects,
    colimits
)
where
    import FiniteCategory.FiniteCategory
    import Diagram.Diagram
    import DiagonalFunctor.DiagonalFunctor
    import CommaCategory.CommaCategory
    import FunctorCategory.FunctorCategory
    import UsualCategories.One
    import Data.Maybe                                  (fromJust)
    import Utils.AssociationList

    -- --------------------------------
    -- Cone related functions and types.
    -- --------------------------------
        
    -- | A `Cone` is a `CommaObject` in the `CommaCategory` (/D/|1_/F/).
    type Cone c1 m1 o1 c2 m2 o2 = CommaObject o2 One (NaturalTransformation c1 m1 o1 c2 m2 o2)
    
    -- | A `ConeMorphism` is a morphism between cones.
    type ConeMorphism c1 m1 o1 c2 m2 o2 = CommaMorphism o2 One m2 One (NaturalTransformation c1 m1 o1 c2 m2 o2)
    
    -- | `ConeCategory` is the type of the cone category, it is a `CommaCategory` (/D/|1_/F/).
    type ConeCategory c1 m1 o1 c2 m2 o2 = CommaCategory c2 m2 o2 One One One (FunctorCategory c1 m1 o1 c2 m2 o2) (NaturalTransformation c1 m1 o1 c2 m2 o2) (Diagram c1 m1 o1 c2 m2 o2)

    -- | Returns the `apex` of a `Cone`.
    apex :: Cone c1 m1 o1 c2 m2 o2 -> o2
    apex = indexSrc

    -- | Returns the `Cone` as a `NaturalTransformation`.
    --
    -- prop> naturalTransformationToCone . coneToNaturalTransformation = id
    -- prop> coneToNaturalTransformation . naturalTransformationToCone = id
    coneToNaturalTransformation :: Cone c1 m1 o1 c2 m2 o2 -> NaturalTransformation c1 m1 o1 c2 m2 o2
    coneToNaturalTransformation = arrow
    
    -- | Returns a `NaturalTransformation` as a `Cone`.
    --
    -- prop> naturalTransformationToCone . coneToNaturalTransformation = id
    -- prop> coneToNaturalTransformation . naturalTransformationToCone = id
    naturalTransformationToCone :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq m1, Eq o1
                                   ,FiniteCategory c2 m2 o2, Morphism m2 o2) =>
                                    NaturalTransformation c1 m1 o1 c2 m2 o2 -> Cone c1 m1 o1 c2 m2 o2
    naturalTransformationToCone nt@NaturalTransformation{srcNT=s,tgtNT=t,component=compo}
        | null (ob (src s)) = error "Cone of a diagram with empty index category"
        | otherwise = CommaObject{indexSrc=(omap s) !-! (head (ob (src s))),indexTgt=One,arrow=nt}

    -- | Constructs the category of cones of a diagram. Objects of the category are `CommaObject` objects with the `apex` of the cone in the `indexSrc` field and the natural transformation in the `arrow` field.
    mkConeCategory :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
                       FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
                       Diagram c1 m1 o1 c2 m2 o2 -> ConeCategory c1 m1 o1 c2 m2 o2
    mkConeCategory diag@Diagram{src=s,tgt=t,omap=om,mmap=fm} =  CommaCategory {leftDiag = diagonalFunct
                                                                             , rightDiag = fromJust $ mkSelect1 (tgt diagonalFunct) diag}
                                                                where
                                                                    diagonalFunct = mkDiagonalFunctor s t

    -- | Returns all cones of a given apex.
    conesOfApex :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
                    FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
                       Diagram c1 m1 o1 c2 m2 o2 -> o2 -> [Cone c1 m1 o1 c2 m2 o2]
    conesOfApex diag apx
        | not $ elem apx (ob (tgt diag)) = error "Trying to construct cones from an apex not in the target category."
        | otherwise = naturalTransformationToCone <$> ar functCat constDiag diag
        where 
            functCat = FunctorCategory (src diag) (tgt diag)
            Just constDiag = mkConstantDiagram (src diag) (tgt diag) apx
            
    -- | Returns limits of a diagram (terminal cones).
    limits :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
               FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
               Diagram c1 m1 o1 c2 m2 o2 -> [Cone c1 m1 o1 c2 m2 o2]
    limits = terminalObjects.mkConeCategory
    
    -- --------------------------------
    -- Cocone related functions and types.
    -- --------------------------------
    

    -- | A `Cocone` is a `CommaObject` in the `CommaCategory` (1_/F/|/D/).
    type Cocone c1 m1 o1 c2 m2 o2 = CommaObject One o2 (NaturalTransformation c1 m1 o1 c2 m2 o2)

    -- | A `CoconeMorphism` is a morphism between cocones.
    type CoconeMorphism c1 m1 o1 c2 m2 o2 = CommaMorphism One o2 One m2 (NaturalTransformation c1 m1 o1 c2 m2 o2)
    
    -- | `CoconeCategory` is the type of the cocone category, it is a `CommaCategory` (1_/F/|/D/).
    type CoconeCategory c1 m1 o1 c2 m2 o2 = CommaCategory One One One c2 m2 o2 (FunctorCategory c1 m1 o1 c2 m2 o2) (NaturalTransformation c1 m1 o1 c2 m2 o2) (Diagram c1 m1 o1 c2 m2 o2)
    
    
    -- | Returns the `nadir` of a `Cocone`.
    nadir :: Cocone c1 m1 o1 c2 m2 o2 -> o2
    nadir = indexTgt

    -- | Returns the `Cocone` as a `NaturalTransformation`.
    coconeToNaturalTransformation :: Cocone c1 m1 o1 c2 m2 o2 -> NaturalTransformation c1 m1 o1 c2 m2 o2
    coconeToNaturalTransformation = arrow
    
    -- | Returns a `NaturalTransformation` as a `Cocone`.
    --
    -- prop> naturalTransformationToCocone . coconeToNaturalTransformation = id
    -- prop> coconeToNaturalTransformation . naturalTransformationToCocone = id
    naturalTransformationToCocone :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq m1, Eq o1
                                     ,FiniteCategory c2 m2 o2, Morphism m2 o2) =>
                                    NaturalTransformation c1 m1 o1 c2 m2 o2 -> Cocone c1 m1 o1 c2 m2 o2
    naturalTransformationToCocone nt@NaturalTransformation{srcNT=s,tgtNT=t,component=compo}
        | null (ob (src t)) = error "Diagram with empty index category"
        | otherwise = CommaObject{indexSrc=One,indexTgt=(omap t) !-! (head (ob (src t))),arrow=nt}
 
    -- | Constructs the category of cocones of a diagram. Objects of the category are `CommaObject` objects with the nadir of the cone in the `indexTgt` field and the natural transformation in the `arrow` field.
    mkCoconeCategory :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
                       FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
                       Diagram c1 m1 o1 c2 m2 o2 -> CoconeCategory c1 m1 o1 c2 m2 o2
    mkCoconeCategory diag@Diagram{src=s,tgt=t,omap=om,mmap=fm} =  CommaCategory {leftDiag = fromJust $ mkSelect1 (tgt diagonalFunct) diag
                                                                               , rightDiag = diagonalFunct}
                                                                where
                                                                    diagonalFunct = mkDiagonalFunctor s t

    -- | Returns all cocones of a given nadir.
    coconesOfNadir :: ( FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
                        FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
                       Diagram c1 m1 o1 c2 m2 o2 -> o2 -> [Cocone c1 m1 o1 c2 m2 o2]
    coconesOfNadir diag nadr
        | not $ elem nadr (ob (tgt diag)) = error "Trying to construct cocones to a nadir not in the target category."
        | otherwise = naturalTransformationToCocone <$> ar functCat diag constDiag
        where 
            functCat = FunctorCategory (src diag) (tgt diag)
            Just constDiag = mkConstantDiagram (src diag) (tgt diag) nadr
            
    -- | Returns colimits of a diagram (initial cocones).
    colimits :: (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq c1, Eq m1, Eq o1,
               FiniteCategory c2 m2 o2, Morphism m2 o2, Eq c2, Eq m2, Eq o2) =>
               Diagram c1 m1 o1 c2 m2 o2 -> [Cocone c1 m1 o1 c2 m2 o2]
    colimits = initialObjects.mkCoconeCategory